%**************************************************************************
% Simulation VS state evolution: DCT matrix

% NOTE: there is an error for the DCT curve of Fig. 1 in
%       http://arxiv.org/pdf/1602.06509v1.pdf
%       Due to a wrong scaling in my original program, the figure is produced
%       with SNR = 53dB, not 50dB
%**************************************************************************
clear all;

rho = 0.4;                      % sparsity level
BETA = linspace(0,1,20);        % parameter that controls divergence

N = 8192;                    
M = fix(N*0.7);   

delta = M/N;

NSIM = 100;
SNRdB = 53;

Iteration = 50;

MSE_simulation = zeros(1,length(BETA));
MSE_SE1 = zeros(1,length(BETA));    % use (5a)    
MSE_SE2 = zeros(1,length(BETA));    % use (11)

for iii = 1:length(BETA)
    
    rand('state',0);
    randn('state',0);
    
    beta = BETA(iii);
    
    % Note: (i) the definition of SNR in Section V, (ii) matrix is column normalized 
    sigma2 = 1/delta * 10^(-SNRdB/10);  
    
    %% State evolution prediction: Use Eqn.(5a)
    % only for Gaussian-Bernoulli prior and soft-thresholding
    v2 = 1;
    threshold = 1;  
    for it = 1:Iteration
        
        % mapping: v^2 --> tau^2
        tau2 = 1/delta * v2 + sigma2;

        % mapping: tau^2 --> v^2
        PDF = @(x) (1-rho) * 2 * normpdf(x,0,sqrt(tau2) ) +...
            rho * 2 * ( normpdf( x,0, sqrt(tau2 + 1/rho) ) );

        div_old = integral(PDF,threshold,Inf,'RelTol',1e-15);

        temp = ( 1 - beta ) * div_old;

        FUN1 = @(x) ( (temp+1) * x ).^2 .* PDF(x);
        FUN2 = @(x) ( - temp * x - threshold ).^2 .* PDF(x);

        Temp1 = integral(FUN1, 0, threshold,'RelTol',1e-15);
        Temp2 = integral(FUN2, threshold, Inf,'RelTol',1e-15);

        MSE = Temp1 + Temp2 - tau2 + 2 * tau2 * beta * div_old;
        
        % new v2
        v2 = MSE;
         
    end    
    MSE_SE1(iii) = MSE;
    
    %% SE, Use Eqn.(11)
    v2 = 1;
    threshold = 1;  
    for it = 1:Iteration
        
        % mapping: v^2 --> tau^2
        tau2 = (1/delta -1) * v2 + sigma2;

        % mapping: tau^2 --> v^2
        PDF = @(x) (1-rho) * 2 * normpdf(x,0,sqrt(tau2) ) +...
            rho * 2 * ( normpdf( x,0, sqrt(tau2 + 1/rho) ) );

        div_old = integral(PDF,threshold,Inf,'RelTol',1e-15);

        temp = ( 1 - beta ) * div_old;

        FUN1 = @(x) ( (temp+1) * x ).^2 .* PDF(x);
        FUN2 = @(x) ( - temp * x - threshold ).^2 .* PDF(x);

        Temp1 = integral(FUN1, 0, threshold,'RelTol',1e-15);
        Temp2 = integral(FUN2, threshold, Inf,'RelTol',1e-15);

        MSE = Temp1 + Temp2 - tau2 + 2 * tau2 * beta * div_old;
        
        % new v2
        v2 = MSE;
         
    end    
    MSE_SE2(iii) = MSE;
    
    %% Simulation
    
    for nsim=1:NSIM   

        index = random('Binomial',1,rho,N,1);
        x = sqrt(1/rho) * randn(N,1);           % normalize variancce
        x = x .* index;
        
        Permulation = randperm(N);
        index_selection = Permulation(1:M);
        
        % column normalized sensing matrix        
        u = dct(x);
        y = sqrt(1/delta) * u(index_selection);
        
        % noise
        y = y + sqrt(sigma2) * randn(M,1);

        % recovery
        
        z_hat = zeros(M,1);
        div = 0;
        x_hat = zeros(N,1);
        
        for it = 1:Iteration

            u_hat = dct(x_hat);
            z_hat = y - sqrt(1/delta) * u_hat(index_selection) + 1/delta * div * z_hat;
            
            r_hat = zeros(N,1);
            r_hat(index_selection) = z_hat;
            
            r_hat = sqrt(1/delta) * idct(r_hat) + x_hat;

            % Soft thresholding
            threshold = 1;
            x_hat_old = max(abs(r_hat) - threshold, 0) .* sign(r_hat);

            % divergence 
            div_old = ( length( find( abs(r_hat) > threshold) ) ) / N; 
            
            % construct a new estimator, update divergence         
            x_hat = x_hat_old - (1 - beta) * div_old * r_hat;
            div = beta * div_old;

        end     
        
        % measure MSE 
        MSE_simulation(iii) = MSE_simulation(iii) + norm(x_hat - x)^2/N; 
        
        % The difference between simulation and state evolution
        mse_simulation = MSE_simulation(iii)/nsim;
        mse_SE1 = MSE_SE1(iii);
        mse_SE2 = MSE_SE2(iii);
        diff1 = abs( mse_simulation - mse_SE1)/mse_simulation;
        diff2 = abs( mse_simulation - mse_SE2)/mse_simulation;
        if(mod(nsim,100)==0)
            fprintf('************************************************************\n');
            fprintf('beta = %d, nsim = %d\n',BETA(iii),nsim);
            fprintf('diff1 = %e, diff2 = %e\n',diff1,diff2);
        end
        
    end

end

MSE_simulation = MSE_simulation/NSIM;
Diff1 = abs(MSE_simulation - MSE_SE1)./MSE_simulation;
Diff2 = abs(MSE_simulation - MSE_SE2)./MSE_simulation;

figure;
plot(BETA,Diff1,'b--',BETA,Diff2,'b-^');
ylabel('relative SE prediction error');
xlabel('\beta');






