%**************************************************************************
% Simulation for AMP + IID Gaussian, Bayes-optimal eta-function
% For Fig. 2 of http://arxiv.org/pdf/1602.06509v1.pdf

% NOTE: A very large matrix is used to verify the accuray of SE
%**************************************************************************

clear all;

rho = 0.1;    % sparsity level

N = 16384;     
M = fix(N * 0.35);  
delta = M/N;

NSIM = 500;
SNRdB = 50;

Iteration = 25;

MSE_simulation = zeros(Iteration,1);

sigma2 = 1/delta * 10^(-SNRdB/10);    
    
for nsim=1:NSIM                                                       
    index = random('Binomial',1,rho,N,1);
    x = sqrt(1/rho) * randn(N,1);
    x = x .* index;

    % column normalized matrix
    A = randn(M,N)/sqrt(M); 
    y = A * x + sqrt(sigma2) * randn(M,1); 
    
    % AMP recovery
    z_hat = zeros(M,1);
    div = 0;
    x_hat = zeros(N,1);

    for it = 1:Iteration
        z_hat = y - A * x_hat + 1/delta * div * z_hat;
        r_hat = A' * z_hat + x_hat;
        
        tau2 = norm(z_hat)^2/M;
        
        % scalar MMSE estimator      
        exponent =  - abs(r_hat).^2 * 1/rho ./ ( tau2 .* (tau2+1/rho));
        exponent = min(exponent, 50);
        exponent = max(exponent, -50);
        exponent = exponent / 2; 
        den = 1 + sqrt( (tau2+1/rho)./tau2) .* (1-rho)/rho .* exp(exponent); 
        C = 1./den;
        
        % conditional mean and variance
        x_hat = ( 1/rho ) ./ ( 1/rho + tau2 ) .* r_hat .* C;
        Var = 1/rho*tau2./(tau2+1/rho) .* C + abs(1/rho./(1/rho+tau2) .* r_hat).^2 .* C .* (1-C);
        
        % divergence
        div = mean(Var)/tau2;
        
        % measure MSE
        MSE_simulation(it) = MSE_simulation(it) + norm(x_hat - x)^2/N;
    end     
     

    if(mod(nsim,1)==0)
        fprintf('************************************************************\n');
        fprintf('snr = %d dB, nsim = %d\n',SNRdB,nsim);
        fprintf('MSE = %e\n',MSE_simulation(Iteration)/nsim);
    end

end

MSE_simulation = MSE_simulation/NSIM;

figure;
semilogy(1:Iteration,MSE_simulation,'r-');
legend('simulation');
xlabel('Iteration');
ylabel('MSE');
