%**************************************************************************
% Simulation for OAMP + LMMSE for IID Gaussian matrix
% For Fig. 2 of http://arxiv.org/pdf/1602.06509v1.pdf

% NOTE: A very large matrix is used to verify the accuray of SE
%       Only for M<N
%**************************************************************************

clear all;

rho = 0.1;    % sparsity level

N = 16384;      % this is very slow. Try a smaller size such as 1024
M = fix(N * 0.35);  
delta = M/N;

NSIM = 500;
SNRdB = 50;

Iteration = 25;

MSE_simulation = zeros(Iteration,1);

sigma2 = 1/delta * 10^(-SNRdB/10); 
tic;
    
for nsim=1:NSIM                                                       
    index = random('Binomial',1,rho,N,1);
    x = sqrt(1/rho) * randn(N,1);
    x = x .* index;

    % column normalized matrix
    A = randn(M,N)/sqrt(M); 
    y = A * x + sqrt(sigma2) * randn(M,1); 
    
    % AMP recovery, for column normalized A
    z_hat = zeros(M,1);
    x_hat = zeros(N,1);         

    for it = 1:Iteration
        
        % estimate variance. 
        z_hat = y - A * x_hat;
        v2 = ( norm(z_hat)^2 - M * sigma2 ) / N;      
        v2 = max(v2, 1e-10);
%         % LMMSE
%         R = ( v2 * A*A' + sigma2 * eye(M) );
%         R = R^(-1);
%         W = v2 *A'*R;
%         W = N/trace(W*A) * W;
%         r_hat = W * z_hat + x_hat;
%         
%         c = 1/delta - 1;
%         tau2 = ( sigma2 + c*v2 + sqrt( ( sigma2 + c*v2 )^2 + 4 * sigma2 * v2 ) ) / 2;

        z_hat2 = ( v2 * A*A' + sigma2 * eye(M) ) \ z_hat;
        x_post = x_hat + v2 * A' * z_hat2;
        V_post = v2 * eye(N) - (v2)^2 * A' * ( v2 * A*A' + sigma2 * eye(M) )^(-1) * A;
        v_post = trace(V_post)/N;
        v_ext = 1/(1/v_post - 1/v2);
        r_hat = v_ext/v_post * x_post - v_ext/v2 * x_hat;
        tau2 = v_ext;
        
        % scalar MMSE estimator      
        exponent =  - abs(r_hat).^2 * 1/rho ./ ( tau2 .* (tau2+1/rho));
        exponent = min(exponent, 50);
        exponent = max(exponent, -50);
        exponent = exponent / 2; 
        den = 1 + sqrt( (tau2+1/rho)./tau2) .* (1-rho)/rho .* exp(exponent); 
        C = 1./den;
        
        % conditional mean and variance
        x_hat_post = ( 1/rho ) ./ ( 1/rho + tau2 ) .* r_hat .* C;
        Var = 1/rho*tau2./(tau2+1/rho) .* C + abs(1/rho./(1/rho+tau2) .* r_hat).^2 .* C .* (1-C);
        
        % divergence
        div = mean(Var)/tau2;
        
        % eliminate divergence
        C = tau2 / (tau2 - mean(Var) );
        x_hat = C * ( x_hat_post - div * r_hat );
        
        % measure MSE, use x_hat_post
        MSE_simulation(it) = MSE_simulation(it) + norm(x_hat_post - x)^2/N;
    end     
     

    if(mod(nsim,1)==0)
        toc;
        fprintf('************************************************************\n');
        fprintf('snr = %d dB, nsim = %d\n',SNRdB,nsim);
        fprintf('MSE = %e\n',MSE_simulation(Iteration)/nsim);
    end

end



MSE_simulation = MSE_simulation/NSIM;

figure;
semilogy(1:Iteration,MSE_simulation,'r-');
legend('simulation');
xlabel('Iteration');
ylabel('MSE');
