%**************************************************************************
% Simulation for OAMP + PINV of
% Fig. 3 of http://arxiv.org/pdf/1602.06509v1.pdf
%**************************************************************************
clear all;

rho = 0.2;      % sparsity level

N = 1000;        % the result in the paper is produced using N=4000        
M = fix(N * 0.5);  
delta = M/N;

NSIM = 100;
SNRdB = 60;

Iteration = 35;

MSE_simulation = zeros(Iteration,1);

sigma2 = 1/delta * 10^(-SNRdB/10); 
tic;
    
%% set up the singular values, column normalization
kappa = 5;

d2 = kappa.^(-2/M*[0:M-1]);
d2 = d2/sum(d2) * N;
d = sqrt(d2)';

%% Simulation
for nsim=1:NSIM                                                       
    index = random('Binomial',1,rho,N,1);
    x = sqrt(1/rho) * randn(N,1);
    x = x .* index;

    % column normalized matrix
    temp = randn(N,N);
    [V_full,~] = qr(temp);
    V = V_full(1:M,:);
    
    temp = randn(M,M);
    [U,~] = qr(temp);

    A = U * diag(d) * V;
    
    y = A * x + sqrt(sigma2) * randn(M,1); 
    
    % AMP recovery, for column normalized A
    z_hat = zeros(M,1);
    x_hat = zeros(N,1); 
    
%     W = 1/delta * A'*(A*A')^(-1);
    W = 1/delta * pinv(A);      % more stable 
    Trace = trace((A*A')^(-1)); 

    for it = 1:Iteration
        
        % estimate variance. 
        z_hat = y - A * x_hat;
        v2 = ( norm(z_hat)^2 - M * sigma2 ) / N;      
        v2 = max(v2, 1e-10);
        
        % PINV
        r_hat = W * z_hat + x_hat;
        tau2 = (1/delta - 1) * v2 + N/M^2 * sigma2 * Trace;
                
        % scalar MMSE estimator      
        exponent =  - abs(r_hat).^2 * 1/rho ./ ( tau2 .* (tau2+1/rho));
        exponent = min(exponent, 50);
        exponent = max(exponent, -50);
        exponent = exponent / 2; 
        den = 1 + sqrt( (tau2+1/rho)./tau2) .* (1-rho)/rho .* exp(exponent); 
        C = 1./den;
        
        % conditional mean and variance
        x_hat_post = ( 1/rho ) ./ ( 1/rho + tau2 ) .* r_hat .* C;
        Var = 1/rho*tau2./(tau2+1/rho) .* C + abs(1/rho./(1/rho+tau2) .* r_hat).^2 .* C .* (1-C);
        
        % divergence
        div = mean(Var)/tau2;
        
        % eliminate divergence
        C = tau2 / (tau2 - mean(Var) );
        x_hat = C * ( x_hat_post - div * r_hat );
        
        % measure MSE, use x_hat_post
        MSE_simulation(it) = MSE_simulation(it) + norm(x_hat_post - x)^2/N;
    end     
     

    if(mod(nsim,1)==0)
        toc;
        fprintf('************************************************************\n');
        fprintf('snr = %d dB, nsim = %d\n',SNRdB,nsim);
        fprintf('MSE = %e\n',MSE_simulation(Iteration)/nsim);
    end

end

MSE_simulation = MSE_simulation/NSIM;

figure;
semilogy(1:Iteration,MSE_simulation,'r-');
legend('simulation');
xlabel('Iteration');
ylabel('MSE');




