%**************************************************************************
% This program reproduces the ADMM-GAMP curve for Fig.4 in
% http://arxiv.org/pdf/1602.06509v1.pdf

% The implementation is not optimized. It has been mentioned in the paper
% https://arxiv.org/abs/1501.01797 
% that the inner LS could be replaced by conjugate gradient solver.
%**************************************************************************
clear all;
clc;

rho = 0.2;                       % sparsity level

N = 500;
M = 250;
delta = M/N;

NSIM = 500;                        % number of simulations. To save time, you can try a much smaller number, say 10.                  

SNRdB = 60;
Kappa = [1 5 10 1e2 1e3 1e4 1e5 1e6];

Iteration = 50;                     % number of iterations  
Ite_ADMM = 50;

MSE_simulation = zeros(Iteration,length(Kappa));

sigma2 = 1/delta * 10^(-SNRdB/10);

for iii = 1:length(Kappa)
    
    rand('state',0);
    randn('state',0); 
    
    %% set up the singular values, column normalization
    kappa = Kappa(iii);

    d2 = kappa.^(-2/M*[0:M-1]);
    d2 = d2/sum(d2) * N;
    d = sqrt(d2)';
    
    % Simulation    

    for nsim = 1:NSIM
        
        pattern = random('Binomial',1,rho,N,1);        
        x = sqrt(1/rho) * randn(N,1);
        x = x .* pattern;
        
        % column normalized matrix
        temp = randn(N,N);
        [V_full,~] = qr(temp);
        V = V_full(1:M,:);

        temp = randn(M,M);
        [U,~] = qr(temp);

        A = U * diag(d) * V;
        
        A_H = A';
        A2 = abs(A).^2;
        A2_H = A2';

        y = A * x + sqrt(sigma2) * randn(M,1);       

        % recovery        
        v_r = ones(N,1); 
        v_p = ones(M,1);
        v_r_tmp = ones(N,1);
        v_p_tmp = ones(M,1);
        v_hat = zeros(N,1);     % 
        q_hat = zeros(N,1);
        s_hat = zeros(M,1);
        p_hat = zeros(M,1);

        theta = 1;
        
        z_hat = zeros(M,1);
        x_hat = zeros(N,1);
        
        v_p_old = ones(M,1);
        v_r_old = ones(N,1);
        
        for it = 1:Iteration

            r_old = zeros(N,1);
            p_old = zeros(N,1);
            % The ADMM inner iteration
            for it_admm = 1:Ite_ADMM

                y_tmp = z_hat + v_p .* s_hat;
                w_tmp = x_hat + v_r .* q_hat;
                y_tilde = y_tmp ./ sqrt(v_p);
                w_tilde = w_tmp ./ v_r;
                A_tilde = diag(1./sqrt(v_p)) * A;                

%                 v_hat = ( A_tilde' * A_tilde + diag(1./v_r) )^(-1) * ( A_tilde' * y_tilde + w_tilde );   
                v_hat = ( A_tilde' * A_tilde + diag(1./v_r) ) \ ( A_tilde' * y_tilde + w_tilde ); 
                %----------------------
                r_hat = v_hat - v_r .* q_hat;
                p_hat = A * v_hat - v_p .* s_hat;                

                %------------------------- for real-valuded -------------------
                exponent =  - abs(r_hat).^2 * 1/rho ./ ( v_r .* (v_r+1/rho));
                exponent = min(exponent, 50);
                exponent = max(exponent, -50);

                exponent = exponent / 2;     % for real signal

                den = 1 + sqrt( (v_r+1/rho)./v_r ) .* (1-rho)/rho .* exp(exponent); 
                C = 1./den;

                x_hat = ( 1/rho ) ./ ( 1/rho + v_r ) .* r_hat .* C;
                VAR = 1/rho*v_r./(v_r+1/rho) .* C + abs(1/rho./(1/rho+v_r) .* r_hat).^2 .* C .* (1-C);
                
                %---------------------- MMSE --------------------
                z_hat = p_hat + v_p ./ ( sigma2 + v_p ) .* ( y - p_hat );  
                q_hat = q_hat + 1./v_r .* ( x_hat - v_hat );
                s_hat = s_hat + 1./v_p .* ( z_hat - A * v_hat );

                %----------------------               
                if(norm(r_old - r_hat)/norm(r_hat) < 1e-4 && norm(p_old - p_hat)/norm(p_hat) < 1e-4)
                    break;
                end
            
                r_old = r_hat; 
                p_old = p_hat;    

            end

            v_x = VAR;
            v_z = sigma2 * v_p ./ ( sigma2 + v_p ); % see GAMP (41)
            v_p_tmp = A2 * v_x;
            v_p_tmp = max(v_p_tmp,eps);
            v_s = ( 1 - v_z./v_p_tmp ) ./ v_p_tmp;
            v_r_tmp = 1./ ( A2_H * v_s );
            v_r_tmp = max(v_r_tmp,eps);
            
            % variance damping
            v_r = 1 ./ (  theta ./ v_r_tmp  + ( 1-theta ) ./ v_r  );
            v_p = 1 ./ (  theta ./ v_p_tmp  + ( 1-theta ) ./ v_p  );
    
            if(theta==1)
                v_r = v_r_tmp;
                v_p = v_p_tmp;
            end
                  
            MSE_simulation(it,iii) = MSE_simulation(it,iii) + norm(x_hat - x)^2/(N);
            
            if(norm(v_p_old - v_p)/norm(v_p) < 1e-4 && norm(v_r_old - v_r)/norm(v_r) < 1e-4)
                for(it2 = it+1:Iteration)
                    MSE_simulation(it2,iii) = MSE_simulation(it2,iii) + norm(x_hat - x)^2/(N);
                end
                break;
            end
            
            v_p_old = v_p;
            v_r_old = v_r;
            
            fprintf('kappa = %e, nsim = %d, it = %d, MSE_sim = %e\n',kappa,nsim,it, MSE_simulation(it,iii)/nsim);
            
        end 
   
    end

    MSE_simulation(:,iii) = MSE_simulation(:,iii)/NSIM;
   
end

figure;
loglog(Kappa,MSE_simulation(Iteration,:),'r-');
legend('ADMM-GAMP');
xlabel('Kappa');
ylabel('MSE');