%**************************************************************************
% Simulation for GAMP_damping for Fig. 4 of
% http://arxiv.org/pdf/1602.06509v1.pdf

% NOTE: 
% (1) this program requires the GAMP package, avaiable at
%     http://sourceforge.net/projects/gampmatlab/
% (2) this program is modified based on an example given in the GAMP
%     package
%**************************************************************************

clear all;
clc;

% Should change to your own directory 
addpath('gampmatlab20150901\code\trunk\code\main');     

%Handle random seed
if verLessThan('matlab','7.14')
  defaultStream = RandStream.getDefaultStream;
else
  defaultStream = RandStream.getGlobalStream;
end;
if 1
    savedState = defaultStream.State;
    save random_state.mat savedState;
else
    load random_state.mat
end
defaultStream.State = savedState;

% Parameters
nx = 500;         % Number of input components (dimension of x)
nz = 250;         % Number of output components (dimension of y)
sparseRat = 0.2;  % fraction of components of x that are non-zero
snr = 60;         % SNR in dB.

% Add by Junjie
Kappa = [1:2:9 10 1e2 1e3 1e4 1e5 1e6];    
MSE_simulation = zeros(1,length(Kappa));
NSIM = 500;   

for iii = 1:length(Kappa)
    
    % Add by Junjie
    kappa = Kappa(iii);
    d2 = kappa.^(-2/nz*[0:nz-1]);
    d2 = d2/sum(d2) * nx;
    d = sqrt(d2)';

    rand('state',0);
    randn('state',0);
    
    for nsim = 1:NSIM
        
        % Create a random sparse vector
        xmean0 = 0;        
        xvar0 = 1;
        x0 = normrnd(xmean0, sqrt(xvar0),nx,1); % a dense Gaussian vector
        x = x0.*(rand(nx,1) < sparseRat);       % insert zeros

        % Add by Junjie
        temp = randn(nx,nx);
        [V_full,~] = qr(temp);
        V = V_full(1:nz,:);

        temp = randn(nz,nz);
        [U,~] = qr(temp);

        A = U * diag(d) * V;

        wvar = 10^(-0.1*snr)*(norm(A*x)^2)/nz;
        w = normrnd(0, sqrt(wvar), nz,1); % AWGN
        z = A*x;
        y = z + w;

        % Test MMV version
        %y = [y,y];

        % Generate input estimation class
        % First, create an estimator for a Gaussian random variable (with no
        % sparsity)
        inputEst0 = AwgnEstimIn(xmean0, xvar0);

        % Then, create an input estimator from inputEst0 corresponding to a random
        % variable x that is zero with probability 1-sparseRat and has the
        % distribution of x in inputEst0 with probability sparseRat.
        inputEst = SparseScaEstim( inputEst0, sparseRat );

        % Output estimation class:  Use the 
        outputEst = AwgnEstimOut(y, wvar);

        % Set GAMP options
        opt = GampOpt();
        opt.legacyOut = false;
        opt.uniformVariance = 0; 
        opt.tol = 1e-5;
        opt.nit = 1000;
        opt.adaptStep = 1;	
        opt.adaptStepBethe = 1;	
        opt.stepWindow =2;	
        opt.stepMax = 1.0;	
        opt.stepMin = 0.01;	% seems important to be >0
        opt.step = opt.stepMax;	
        opt.stepIncr = 1.1;
        opt.stepDecr = 0.5;

        opt.varNorm = 0;	
        opt.xvarMin = 0;	
        opt.pvarMin = 0;	
        opt.pvarStep = 1;	
        opt.rvarStep = 1;	

        % Run GAMP without mean removal
        opt.removeMean = 0;
        estFin = gampEst(inputEst, outputEst, A, opt);
        xhat_naive = estFin.xhat(:,1);% for MMV version

        nmseGAMP_naive = norm(x-xhat_naive)^2/norm(x)^2;
        
        % Add by Junjie
        MSE_simulation(iii) = MSE_simulation(iii) + nmseGAMP_naive;        
        if(mod(nsim,1)==0)
            fprintf('======================================================\n');
            fprintf('snr = %d dB, kappa = %e, nsim = %d\n',snr,kappa,nsim);
            fprintf('MSE_sim = %e\n',MSE_simulation(iii)/nsim);
        end
    
    end    
    
    MSE_simulation(iii) = MSE_simulation(iii)/NSIM; 
end

figure;
loglog(Kappa,MSE_simulation,'r-');
legend('GAMP-damping');
xlabel('Kappa');
ylabel('MSE');