%**************************************************************************
% Simulation for OAMP + PINV For Fig. 4 of
% http://arxiv.org/pdf/1602.06509v1.pdf
%**************************************************************************

clear all;

rho = 0.2;      % sparsity level

N = 500;        % the result in the paper is produced using N=4000        
M = fix(N * 0.5);  
delta = M/N;

NSIM = 500;
SNRdB = 60;

Iteration = 50;

Kappa = [1:1:1e1 1e2 1e3 1e4 1e5 1e6];
% Kappa = 1e6;
MSE_simulation = zeros(length(Kappa),1);

sigma2 = 1/delta * 10^(-SNRdB/10); 

%% Simulation
for iii = 1:length(Kappa)
    
    %% set up the singular values, column normalization
    kappa = Kappa(iii);

    d2 = kappa.^(-2/M*[0:M-1]);
    d2 = d2/sum(d2) * N;
    d = sqrt(d2)';
    
    for nsim=1:NSIM                                                       
        index = random('Binomial',1,rho,N,1);
        x = sqrt(1/rho) * randn(N,1);
        x = x .* index;

        % column normalized matrix
        temp = randn(N,N);
        [V_full,~] = qr(temp);
        V = V_full(1:M,:);

        temp = randn(M,M);
        [U,~] = qr(temp);

        A = U * diag(d) * V;

        y = A * x + sqrt(sigma2) * randn(M,1); 

        % AMP recovery, for column normalized A
        z_hat = zeros(M,1);
        x_hat = zeros(N,1);         

        stop = 0;
        
        x_hat_post_old = zeros(N,1);
        
        it = 0;
        
%         W = 1/delta * A'*(A*A')^(-1);
        W = 1/delta * pinv(A);  % more stable
        Trace = trace((A*A')^(-1));
        
        while (~stop) 
            it = it + 1;
            
            % estimate variance. 
            z_hat = y - A * x_hat;
            v2 = ( norm(z_hat)^2 - M * sigma2 ) / N;      
            v2 = max(v2, 1e-10);

            % PINV
            r_hat = W * z_hat + x_hat;
            tau2 = (1/delta - 1) * v2 + N/M^2 * sigma2 * Trace;

            % scalar MMSE estimator      
            exponent =  - abs(r_hat).^2 * 1/rho ./ ( tau2 .* (tau2+1/rho));
            exponent = min(exponent, 50);
            exponent = max(exponent, -50);
            exponent = exponent / 2; 
            den = 1 + sqrt( (tau2+1/rho)./tau2) .* (1-rho)/rho .* exp(exponent); 
            C = 1./den;

            % conditional mean and variance
            x_hat_post = ( 1/rho ) ./ ( 1/rho + tau2 ) .* r_hat .* C;
            Var = 1/rho*tau2./(tau2+1/rho) .* C + abs(1/rho./(1/rho+tau2) .* r_hat).^2 .* C .* (1-C);

            % divergence
            div = mean(Var)/tau2;

            % eliminate divergence
            C = tau2 / (tau2 - mean(Var) );
            x_hat = C * ( x_hat_post - div * r_hat );
            
            if( norm(x_hat_post-x_hat_post_old)^2/norm(x_hat_post)^2 < 1e-10 || it > Iteration )
                stop = 1;
            end
            
            x_hat_post_old = x_hat_post;
            
        end     
        
        % measure MSE, use x_hat_post
        
        MSE_simulation(iii) = MSE_simulation(iii) + norm(x_hat_post - x)^2/N;

        if(mod(nsim,1)==0)
            fprintf('************************************************************\n');
            fprintf('kappa = %e, nsim = %d\n',kappa,nsim);
            fprintf('MSE = %e\n',MSE_simulation(iii)/nsim);
        end
    end

end

MSE_simulation = MSE_simulation/NSIM;

figure;
loglog(Kappa,MSE_simulation,'r-');
legend('OAMP-PINV');
xlabel('Kappa');
ylabel('MSE');




